local bit = require 'bit'
local ffi = require 'ffi'
local sample_clock = require 'sample_clock'

local bit_band = bit.band
local bit_bor = bit.bor
local bit_rshift = bit.rshift
local ffi_cast = ffi.cast
local ffi_copy = ffi.copy
local ffi_new = ffi.new
local ffi_sizeof = ffi.sizeof
local ffi_string = ffi.string

local htonl = ffi.abi('le') and bit.bswap or bit.tobit

ffi.cdef[[
typedef struct __attribute((packed)) message_can {
    int64_t timestamp[1];
    uint8_t value[13];
} message_can;
]]

local hex = ffi_new('uint8_t[16]', '0123456789ABCDEF')
local c_type = 'message_can'

local M = {}

function M.id(msg)
    return htonl((ffi_cast('uint32_t *', msg.value))[0])
end

function M.ff(msg)
    return msg.value[12]
end

function M.dlc(msg)
    return bit_band(msg:ff(), 0x0F)
end

function M.data_length(msg)
    local dlc = msg:dlc()
    return dlc > 8 and 8 or dlc
end

function M.byte_length(msg)
    if msg:is_heartbeat()
        then return 0
    end
    return msg:data_length() + 4
end

function M.is_tx(msg)
    return bit_band(msg:ff(), 0x10) ~= 0
end

function M.rtr(msg)
    return bit_band(msg:ff(), 0x40) ~= 0
end

function M.is_heartbeat(msg)
    return bit_band(msg:ff(), 0x20) ~= 0
end

function M.set_heartbeat(msg, setting)
    msg.value[12] = bit_bor(bit_band(msg:ff(), 0xfd), setting and 0x20 or 0)
end

function M.dump(msg)
    local size = ffi_sizeof(c_type)
    local buf = ffi_new('uint8_t[?]', size * 4, 0x20)
    local pos = 0

    local cp = ffi_cast('uint8_t *', msg)

    for i = 0, size - 1 do
        buf[pos] = hex[bit_band(0x0F, bit_rshift(cp[i], 4))]
        buf[pos + 1] = hex[bit_band(0x0F, cp[i])]
        pos = pos + 2
        if i == 7 or i >= 11 then
            pos = pos + 1
        end
    end

    return string.format("%.9f %s (%p %s)",
        sample_clock.secs(msg.timestamp[0]), msg, msg, ffi_string(buf, pos))
end

function M.__len(msg)
    return msg:byte_length()
end

function M.__tostring(msg)
    local buf = ffi_new('uint8_t[64]', 0x20) -- KLUDGE too big maybe
    local pos = 0

    ffi_copy(buf + pos, msg:is_tx() and 'TX' or 'RX', 2)
    pos = pos + 3

    for i = 0, 3 do
        buf[pos] = hex[bit_band(0x0F, bit_rshift(msg.value[i], 4))]
        buf[pos + 1] = hex[bit_band(0x0F, msg.value[i])]
        pos = pos + 2
    end

    for i = 4, 4 + msg:data_length() - 1 do
        buf[pos + 1] = hex[bit_band(0x0F, bit_rshift(msg.value[i], 4))]
        buf[pos + 2] = hex[bit_band(0x0F, msg.value[i])]
        pos = pos + 3
    end

    return ffi_string(buf, pos)
end

M.__index = M

local function init(self, ...)
    local args = {...}
    if #args == 1 and type(args[1]) == 'string' and args[1]:match('^0x') then
        local h = {}
        assert(#args[1] % 2 == 0)
        args[1]:gsub('^0x', ''):gsub('(..)',
            function(m) h[#h + 1] = tonumber(m, 16) end)
        args = h
    end
    assert(#args >= 0 and #args <= 13)
    local msg = self.new({0}, args)
    if #args < 13 then
        msg.value[12] = math.max(#args - 4, 0)
    end
    return msg
end

return {
    new = ffi.metatype(c_type, M),
    c_type_name = c_type,
    c_type = ffi.typeof(c_type),
    ptr_type = ffi.typeof(c_type .. '*'),
    size = ffi.sizeof(c_type),
    init = init,
}
