--
-- Copyright (C) 2015 Hottinger Baldwin Measurements, Inc.
--

local channel_proxy = require 'channel_proxy'
local task_chain    = require 'task_chain'
local log           = require 'log'
local cjson         = require 'cjson.safe'
local __            = require '__'
local sys_notify    = require 'sys_notify'

local mxChannelProxy = class('mxChannelProxy', channel_proxy)

-- The MX471B will send us data values at least at this interval, in
-- seconds. If this time elapses without the MX471B receiving a new
-- CAN message, it will send us a "no data" value.
local MIN_REPEAT = 1

-- The XML-escaped preamble statement of an XML document.
local XML_PREAMBLE_ESCAPED = '&lt;?xml version=&quot;1.0&quot;' ..
    ' encoding=&quot;UTF-8&quot;?&gt;'

function mxChannelProxy:initialize(mxDevice, channel_num)

    self.mxdev              = mxDevice
    self.proxy_type         = 'mxChannelProxy'
    self.index              = channel_num
    self.queued_channels    = {}

    self.xpath_root = '/AdditionalConnector'

    self.spec = mxDevice.mx.control:get_spec(channel_num)
    self.spec.name = string.format('%s_ch_%d', mxDevice.name:gsub('-', '_'), channel_num)

    mxDevice.no_stream = true

end

function mxChannelProxy:available_channels(hostname)
    self.spec.connector = string.format('@%s', self.proxy_hostname)
    self.spec.category =
        string.format('sink/hbm_%s', self.mxdev.category_name)
    local cnum, base = self.proxy_hostname:match('connector(%d+)%.(mx.+)$')
    self.spec.description = string.format("%s Connector %d", base:upper(), cnum)

    return {self.spec}
end

function mxChannelProxy:start_comet_stream(w, chans, sink)
end

function mxChannelProxy:start_channel_stream(w, spec, md_sink, sink)
    self.mxdev.start_channel_list = self.mxdev.start_channel_list or {}
    self.mxdev.start_channel_list[spec._test_id] =
        self.mxdev.start_channel_list[spec._test_id] or {}

    local num = 0
    for k, v in pairs(self.mxdev.start_channel_list) do
        num = num + 1
        if num > 1 then
            local msg = __'Can not start more than one test/MX device at a time'
            log:log(log.ERR, msg)

            for tid, t in pairs(self.mxdev.start_channel_list) do
                for _, c in ipairs(t) do c.md_sink{error_message = msg} end
            end

            md_sink{error_message = msg}
            return
        end
    end

    local start_channel_list = self.mxdev.start_channel_list[spec._test_id]

    -- Get the input module UUID and the reference
    local channel_num, uuid
    if spec.input.connector and spec.input.signal then
        -- Extract UUID and connector number from connector field
        channel_num, uuid =
            spec.input.connector:match('@connector(%d+)%.mx%d+[bc]%-(.+)$')
        spec.input.reference = string.format("AnalogIn_Connector%s.Signal%s",
            tostring(channel_num), tostring(spec.input.signal or 2))
    elseif not spec.input.module or not spec.input.reference then
        md_sink{
            error_message =
                __'Module and reference required when connector is not given',
        }
        return
    else
        -- Extract UUID from module filed.  Channel number is not extracted
        -- because it is embedded in the reference field.
        uuid = spec.input.module:match('^mx%d+[bc]%-(.+)$')
        if not module then
            md_sink{
                error_message = __'Unable to extract UUID from module',
            }
            return
        end

        -- If no signal specified, add it
        if not spec.input.reference:find('%.Signal%d$') then
            spec.input.reference = spec.input.reference .. '.Signal1'
        end
    end

    -- Set the UUID
    spec.input.uuid = uuid:upper()

    start_channel_list[#start_channel_list + 1] = {
        w = w,
        spec = spec,
        md_sink = md_sink,
        sink = sink,
        chan_obj = self,
    }

    self.mxdev.start_channel_list[spec._test_id] = start_channel_list

    if #start_channel_list == spec._channel_count then
        self:_start_channels(start_channel_list)
    end

    w.cleanup = function()
            self.mxdev.start_channel_list[spec._test_id] = nil
            if #self.mxdev.start_channel_list == 0 then
                -- Disable the outputs for this device
                for i = 1, 8 do self:disable_output(i) end
            end
        end
end

local function tag(name, value)
    return '&lt;' .. name .. '&gt;' .. value .. '&lt;/' .. name .. '&gt;'
end

local function open_tag(name)
    return '&lt;' .. name .. '&gt;'
end

local function close_tag(name)
    return '&lt;/' .. name .. '&gt;'
end

local function empty_tag(name)
    return '&lt;' .. name .. ' /&gt;'
end

-- Function used to configure all of the channels
function mxChannelProxy:_start_channels(start_channel_info_list)
    local filter_code = {
        none = 0,
        bessel_iir = 1,
        butterworth_iir = 2,
        linear_phase_fir = 3,
        butterworth_fir = 3,
    }
    local tasks = task_chain:new()

    -- Process all of the channels
    for _, info in ipairs(start_channel_info_list) do
        local spec = info.spec

        -- Get the channel number from the connector
        local cnum = spec.connector:match('^@connector(%d+)%.mx878')

        -- If the filter is not set, default it to off
        if not spec.digital_filter_type then
            spec.digital_filter_type = 'none'
        elseif not filter_code[spec.digital_filter_type] then
            -- Error
            info.md_sink{
                error_message = string.format(
                    __"Unknown digital filter type '%s'",
                    spec.digital_filter_type)
            }
            return
        end

        -- Sanity check for sensor scaling mode
        if spec.calibration[1].mode ~= 'defined_two_point' and
            spec.calibration[1].mode ~= 'defined_slope_intercept' then
            -- ERROR
            info.md_sink{
                error_message = string.format(
                    __"Unknown scaling mode '%s'", spec.calibration[1].mode)
            }
            return
        end
        if spec.calibration[1].purpose ~= 'output_scaling' then
            -- ERROR
            info.md_sink{
                error_message = string.format(
                    __"Unknown scaling purpose '%s'",
                    spec.calibration[1].purpose)
            }
            return
        end
        local xml = {}

        -- XML for the "Input" section
        local module_reference = string.gsub(
            tostring(spec.input.uuid), '%..*$', '')
        xml[#xml + 1] =
            XML_PREAMBLE_ESCAPED ..
            open_tag('Input address=&quot;1&quot;') ..
            tag('ModuleReference',
                string.format("UUID=%s", module_reference)) ..
            tag('SignalReference', tostring(spec.input.reference)) ..
            close_tag('Input')

        -- XML for the "Scaling" section
        xml[#xml + 1] = XML_PREAMBLE_ESCAPED .. open_tag('Scaling')
        if spec.calibration[1].mode == 'defined_slope_intercept' then
            xml[#xml] = xml[#xml] ..
            open_tag('Scalingtype choice=&quot;Gain&quot;') ..
            open_tag('Gain') .. 
            tag('m', tostring(spec.calibration[1].slope)) ..
            tag('b', tostring(spec.calibration[1].intercept)) ..
            close_tag('Gain') ..
            close_tag('Scalingtype')
        elseif spec.calibration1[1].mode == 'defined_two_point' then
            xml[#xml] = xml[#xml] ..
            open_tag('Scalingtype choice=&quot;LinTable&quot;') ..
            open_tag('Point address=&quot;1&quot;') ..
            tag('X', tostring(spec.calibration[1].electrical[1])) ..
            tag('Y', tostring(spec.calibration[1].physical[1])) ..
            close_tag('Point') ..
            open_tag('Point address=&quot;2&quot;') ..
            tag('X', tostring(spec.calibration[1].electrical[2])) ..
            tag('Y', tostring(spec.calibration[1].physical[2])) ..
            close_tag('Point') ..
            close_tag('LinTable') .. 
            close_tag('Scalingtype')
        end
        xml[#xml] = xml[#xml] .. close_tag('Scaling')

        -- XML for the "FilterType" section
        xml[#xml + 1] = XML_PREAMBLE_ESCAPED
        if spec.digital_filter_type == 'none' then
            xml[#xml] = xml[#xml] ..
            open_tag('FilterType choice=&quot;Off&quot;') ..
            empty_tag('Off')
        else
            xml[#xml] = xml[#xml] ..
            open_tag('FilterType choice=&quot;Lowpass&quot;') ..
            tag('FilterCharact', filter_code[spec.digital_filter_type]) ..
            tag('FiltFreq',
                tostring(spec.digital_filter_attenuation_frequency))
        end
        xml[#xml] = xml[#xml] .. close_tag('FilterType')

        -- Base XPATH for the above stuff
        local xpath = 'AdditionalConnector' ..
            string.format('/Connector[@address=&quot;%d&quot;]', cnum) ..
            '/Mode[@choice=&quot;AnalogOut&quot;]/AnalogOut'

        -- Add a "task" per XML blob
        for _, x in ipairs(xml) do
            tasks:add(function(on_success, on_error)
                return self.mxdev.mx.control:set_view_data(8, xpath, x,
                    function(results)
                        if results.result == 'success' then
                            if on_success then return on_success() end
                        else
                            info.md_sink{error_message = results.msg}
                            if on_error then
                                return on_error(results.msg)
                            end
                        end
                    end)
                end)
        end
    end

    tasks:run(nil, function (err)
        local sn = sys_notify:new()
        sn:critical_error('MX878 channel setup failed: ' .. err, true)
    end)

end

-- Function used to disable an analog output
function mxChannelProxy:disable_output(cnum)
    local tasks = task_chain:new()

    local xml = {}

    -- XML for the "Mode" section
    xml[#xml + 1] =
        XML_PREAMBLE_ESCAPED ..
        open_tag('Mode choice=&quot;Off&quot;') ..
        empty_tag('Off') ..
        close_tag('Mode')

    -- Base XPATH for the above stuff
    local xpath = 'AdditionalConnector' ..
        string.format('/Connector[@address=&quot;%d&quot;]', cnum)

    -- Add a "task" per XML blob
    for _, x in ipairs(xml) do
        tasks:add(function(on_success, on_error)
            return self.mxdev.mx.control:set_view_data(8, xpath, x,
                function(results)
                    if results.result == 'success' then
                        if on_success then return on_success() end
                    else
                        if on_error then
                            return on_error(results.msg)
                        end
                    end
                end)
            end)
    end
    tasks:run()
end

return mxChannelProxy
