require 'strict'

-- over_range_detector.lua - Implements the Over Range Detector
-- computed channel by wrapping the functions in over_range_detector.c

local dbg = require 'dbg'
local ffi = require 'ffi'
local C = ffi.C
local config = require 'system_config'
local frame = require 'test_engine/frame'
local pipe = require 'test_engine/pipe'
local rational = require 'rational'
local utils = require 'test_engine/test_eng_utils'

local libte = ffi.load(config.test_engine_library)

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local Pipe_attach_pipe = pipe.attach_pipe
local Pipe_new_pipe_list = pipe.new_pipe_list

-- define structures for interface to C function
-- and C function prototypes
-- NOTE:  These MUST match the contents of cc_over_range_detector.h !!!
ffi.cdef[[
    typedef struct
    {
        char           name[100];
        double         sample_rate;
        uint8_t        mode;
        uint8_t        notify;
        uint16_t       element_size;
        double         alarm_min;
        double         alarm_max;
        struct pipe *  in_pipe;
    } Over_Rng_Detect_chan_init_info;

    typedef struct
    {
        char           name[100];
        double         frame_rate;
        uint32_t       num_channels;
        struct pipe ** pipe_list;
    } Over_Rng_Detect_Node_Profile;

    typedef struct
    {
        uint32_t  nbr;
        uint32_t  exnbr;
        char *    strs[50];  // MUST be same size as in cc_over_range_detector.h
    } Over_Rng_Detect_req;

    uint32_t Over_Rng_Detect_init(Over_Rng_Detect_Node_Profile * profile,
                                  Over_Rng_Detect_chan_init_info * chan_info);

    Over_Rng_Detect_req * Over_Rng_Detect_run(uint32_t module_id);
    void Over_Rng_Detect_stop(uint32_t module_id);
]]

-- These values must match the #defines of the same name
-- in cc_over_range_detector.h !!!
local MODE_NORMAL   = 1
local MODE_MAX_ONLY = 2
local MODE_MIN_ONLY = 3

local mt = {}

function mt:dump()
    dbg('tot_count=%lld', self.tot_count)
end

-- Performs any needed post instantiation initialization.
-- Parameter:  SIE writer - NOT USED by this module!
-- Return values:  true (does nothing, for this module)
function mt:setup(sie_writer)
    return true
end

-- Simply call the C run function - the C code should already
-- be set up with the pipe information via a call to Over_Rng_Detect_init.
function mt:run()
    local notice = nil
    local nbr = 0
    local exnbr = 0
    local ret_val = libte.Over_Rng_Detect_run(self.module_id)
    if ret_val ~= nil then
        local ret_struct = ffi.new("Over_Rng_Detect_req *", ret_val)
        local ret_strs = ffi.new("char **", ret_struct[0].strs)
        nbr = ret_struct[0].nbr
        notice = {}
        for i = 1, nbr do
            notice[i] = ffi.string(ret_strs[i - 1])
        end
        exnbr = ret_struct[0].exnbr
    end
    return notice, nbr, exnbr
end


local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> the index of the first sie buffer used by
--       this instance - this module only uses it to set
--       'sie_collect_idx' which is used by run-test.lua
-- Return values:
--   the new instance object
--   the number of output pipe_lists available
function m:new(stanza, node_list, first_sie_index, meta_list)
    assert(stanza, "Missing parameter 'stanza'")
    local over_channel = setmetatable({}, {__index = mt})
    over_channel.sie_collect_idx = first_sie_index
    utils.set_common_meta(over_channel, stanza)
    -- validate needed metadata, and create input/output pipe
    -- structures
    local profile = ffi.new("Over_Rng_Detect_Node_Profile[1]");
    local name_len = #stanza.name < ffi.sizeof(profile[0].name) and
                     #stanza.name or ffi.sizeof(profile[0].name) - 1
    ffi.fill(profile[0].name, ffi.sizeof(profile[0].name))
    ffi.copy(profile[0].name, stanza.name, name_len)
    -- The actual frame rate is provided by this channel's stanza
    -- as 'sample_rate'; relying on the actual frame rate specified
    -- in the summary section of the .sxr file is inaccurate in the
    -- case of the "Classic" rate domain.
    local frame_rate = rational:new(stanza.sample_rate):tonumbers()
    assert(frame_rate and frame_rate > 0,
           string.format("'%s' sample rate is missing or invalid (%s)",
                         tostring(stanza.name), tostring(frame_rate)))
    profile[0].frame_rate = frame_rate
    assert(node_list, "Missing parameter 'node_list'")
    local input_channels = stanza.input_channels or
        stanza.summary.input_channels or {}
    assert(#input_channels >= 1, "Must have at least one input channel")
    profile[0].num_channels = #input_channels
    local input_chans =
        ffi.new("Over_Rng_Detect_chan_init_info[?]", #input_channels)
    -- get channel information for input channels
    for i = 0, #input_channels - 1 do
        -- create info entry for this input channel
        local node_meta = meta_list[input_channels[i+1]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
               tostring(input_channels[i+1]) .. "\" not found")
        --     name
        assert(node_meta.name, "Missing channel name")
        local name_len = (#node_meta.name < 100) and #node_meta.name or 99
        ffi.copy(input_chans[i].name, node_meta.name, name_len)
        input_chans[i].name[name_len] = 0
        local sample_rate = rational:new(node_meta.sample_rate):tonumbers()
        --     sample_rate
        assert(sample_rate and sample_rate > 0,
               string.format("sample rate is missing or invalid (%s) for '%s'",
                             tostring(sample_rate), tostring(node_meta.name)))
        input_chans[i].sample_rate = sample_rate
        --     data_type
        assert(node_meta.data_type,
                string.format("Missing data type for '%s'",
                    tostring(node_meta.name)))
        local element_size = utils.sample_size(node_meta.data_type)
        assert(element_size <= 8, string.format("Invalid element size (%s)" ..
                " for '%s'", tostring(element_size), tostring(node_meta.name)))
        input_chans[i].element_size = element_size
        --     alarm range(s)
        assert(node_meta.range_alerts and (node_meta.range_alerts.alarm_max or
                node_meta.range_alerts.alarm_min),
                string.format("Missing both alarm min and alarm max for '%s'",
                    tostring(node_meta.name)))
        local negative, range, percent, maxstr
        if node_meta.range_alerts.alarm_max then
            negative, maxstr, percent =
                tostring(node_meta.range_alerts.alarm_max):match("(%-?)([%d%.]+)(%%?)")
            range = rational:new(maxstr):tonumbers()
            if negative == '-' then
                range = -1.0 * range
            end
            assert(range, string.format("Invalid maximum alarm value (%s) for '%s'",
                    tostring(node_meta.range_alerts.alarm_max), tostring(node_meta.name)))
            if percent == "%" then
                percent = rational:new(range):tonumbers()
                assert(percent and percent > 0,
                       string.format("Invalid alarm percentage (%s) for '%s'",
                        tostring(percent), tostring(node_meta.name)))
                assert(node_meta.range_max,
                        string.format("Missing range maximum for '%s'",
                            tostring(node_meta.name)))
                local rmax = rational:new(node_meta.range_max):tonumbers()
                assert(rmax, string.format("Invalid range maximum (%s) for '%s'",
                        tostring(rmax), tostring(node_meta.name)))
                input_chans[i].alarm_max = rmax * percent / 100
            else
                input_chans[i].alarm_max = range
            end
        end
        local minstr
        if node_meta.range_alerts.alarm_min then
            negative, minstr, percent =
                tostring(node_meta.range_alerts.alarm_min):match("(%-?)([%d%.]+)(%%?)")
            range = rational:new(minstr):tonumbers()
            if negative == '-' then
                range = -1.0 * range
            end
            assert(range, string.format("Invalid minimum alarm value (%s) for '%s'",
                    tostring(node_meta.range_alerts.alarm_min), tostring(node_meta.name)))
            if percent == "%" then
                percent = rational:new(range):tonumbers()
                assert(percent and percent > 0,
                       string.format("Invalid alarm percentage (%s) for '%s'",
                        tostring(percent), tostring(node_meta.name)))
                assert(node_meta.range_min,
                        string.format("Missing range minimum for '%s'",
                            tostring(node_meta.name)))
                local rmin = rational:new(node_meta.range_min):tonumbers()
                assert(rmin, string.format("Invalid range minimum (%s) for '%s'",
                        tostring(rmin), tostring(node_meta.name)))
                input_chans[i].alarm_min = rmin * percent / 100
            else
                input_chans[i].alarm_min = range
            end
        end
        assert(minstr or maxstr,
            "Must have at least one of 'alarm minimum' and 'alarm maximum'")
        --     mode, derived from existence of alarm min/max
        input_chans[i].mode =
                (minstr and maxstr and MODE_NORMAL) or
                (minstr and MODE_MIN_ONLY) or MODE_MAX_ONLY
        --     set flag indicating whether to send notifications
        input_chans[i].notify = stanza.enable_notifications and
            (tostring(stanza.enable_notifications) == "true") and 1 or 0
        --     input pipe
        local input_node = node_list[input_channels[i+1]]
        assert(input_node, "Node for Input Channel \"" ..
               tostring(input_channels[i+1]) .. "\" not found")
        input_chans[i].in_pipe = Pipe_attach_pipe(input_node.pipe_list)
    end

    over_channel.pipe_list = Pipe_new_pipe_list()
    assert(over_channel.pipe_list and over_channel.pipe_list ~= nil,
        "failed to create pipe list")
    profile[0].pipe_list = over_channel.pipe_list

    -- initialize the C code, which does the actual processing for
    -- this computed channel
    local mod_id = libte.Over_Rng_Detect_init(profile, input_chans)
    assert(mod_id >= 0, "Failed to create Over Range Detector module")
    over_channel.module_id = mod_id

    -- used for frames' time stamps (just a sample count for
    -- this module)
    over_channel.tot_count = 0

    return over_channel, 1
end

return m
