require 'strict'

-- time_history.lua - Implements the Triggered Time History
-- data mode.
--
-- Implementation note:
--     Data modes in general handle setting up their own trigger
--     mechanisms, where such are called for.  For those (such as
--     this one) which use the "generic" triggering mechanism a
--     trigger.lua module will be implemented.  The trigger module
--     will be given all this data mode instance's input channels,
--     along with the trigger channel, and the trigger module will
--     then provide this data mode instance with an input pipe
--     for each input channel.  This data mode instance will then
--     simply write each input pipe's frames to the associated
--     sie buffer pipe.

local cjson = require 'cjson'
local ffi = require 'ffi'
local frame = require 'test_engine/frame'
local __ = require '__'
local log = require 'log'
local pipe = require 'test_engine/pipe'
local rational = require 'rational'
local sie_schema = require 'test_engine/sie_schema'
local sie_utils = require 'test_engine/sie_utils'
local utils = require 'test_engine/test_eng_utils'

local ffi_cast = ffi.cast
local ffi_sizeof = ffi.sizeof

local Frame_create = frame.create
local Frame_delete = frame.delete
local Frame_tail   = frame.tail
local Pipe_attach_pipe   = pipe.attach_pipe
local Pipe_has_frames    = pipe.has_frames
local Pipe_get_frame     = pipe.get_frame
local Pipe_new_pipe_list = pipe.new_pipe_list

local mt = {}

-- Performs any needed post instantiation initialization.
-- Parameter:
--     'sie_writer'  SIE writer
-- Returns:  true
function mt:setup(sie_writer)
    self.sw = sie_writer
    self:sie_xml()
    return true
end

-- Creates the SIE buffers (pipes and XML) required.
-- There will be one channel description per input channel.
function mt:sie_xml()
    for i = 1, #self.chan_info do
        local info = self.chan_info[i]
        -- write XML for this SIE channel
        info.stanza.description = self.stanza.description or ""
        info.stanza.datamode_name = self.name
        if self.prefix_dm_name then
            info.stanza.sie_name = self.name .. '@' .. info.name
        else
            info.stanza.sie_name = info.name
        end
        info.stanza.somat_input_channel = info.stanza.name
        -- the following line makes the schema info reflect the
        -- input channel's data bit size (?should this be different?)
        self.stanza.nbr_data_bits = 
            tonumber(info.data_type:match('^.-(%d+)$'))
        assert(self.stanza.nbr_data_bits and self.stanza.nbr_data_bits > 0,
            "Missing data bit size")
        local schema_name = sie_schema.get_dm_schema_name(self.stanza.category)
        local schema_info = sie_schema.get_schema_info(schema_name, self.stanza)
        if self.stanza.trigger_mode == "always_on" then
            schema_info.dims[1]['core:label'] = "Time"
            schema_info.dims[1]['core:units'] = "Seconds"
        else
            schema_info.dims[1]['core:label'] = "Collection Time"
            schema_info.dims[1]['core:units'] = "Seconds"
        end
        schema_info.dims[2]['core:label'] = info.name
        schema_info.dims[2]['core:units'] = info.stanza.output_units
        schema_info['core:uuid'] = self.stanza.id .. '/' .. info.stanza.id

        info.stanza.sie_buf_info = info.sie_buf_info
        local md = sie_utils.write_sie_channel(self.sw, schema_info,
            info.stanza)
        -- Attach one end of pipe to this module for this input node
        local p = Pipe_attach_pipe(info.pipe_list)
        -- Create the SIE buffer and attach the pipe to it
        local size = utils.sie_buf_size(self, info.stanza)
        self.sw:new_buffer(info.sie_buf_info.idx, info.sie_buf_info.group,
            size, ffi_sizeof(info.data_type), 28, p)
        self.sw.bufs[info.sie_buf_info.idx].ch_id = info.sie_buf_info.id
    end
end

-- Gets a frame from each input pipe, storing the frames in
-- the input pipe's 'self.chan_info[<pipe index>].in_frame' entry.
-- Returns:  true if a full set of frames is acquired,
--           false otherwise
function mt:get_frames()
    -- check that each input pipe has a frame, so we don't
    -- "get" a frame from any pipe which we can't use
    local got_frames = #self.chan_info > 0
    if not got_frames then
        return false  -- short circuit return!
    end

    for i = 1, #self.chan_info do
        got_frames = Pipe_has_frames(self.chan_info[i].in_pipe)
        if not got_frames then
            break  -- don't have a frame from this input pipe
        end
    end
    if got_frames then
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame =
                Pipe_get_frame(self.chan_info[i].in_pipe)
        end
    else
        -- ensure we don't accidentally use old data
        for i = 1, #self.chan_info do
            self.chan_info[i].in_frame = nil
        end
    end
    return got_frames
end

-- Processes incoming frames, producing outgoing frames.
function mt:run()
    if self.trigger_module then
        -- trigger node feeds this node
        self.trigger_module:run()
    end

    -- attempt to get a full set of input frames (one from each
    -- input pipe)
    local have_all_frames = self:get_frames()
    if not have_all_frames then
        -- can't process any input frames yet
        return
    end

    -- Process each frame in input pipe(s).
    while have_all_frames do
        -- NOTE:  We rely on all input pipes using the same sample rate!
        --        It doesn't matter which input frame we use for
        --        obtaining frame count.
        local count = self.chan_info[1].in_frame.count

         -- get input frames data "pointers"
        for i = 1, #self.chan_info do
            local infr = self.chan_info[i].in_frame
            local sample_type = utils.frame_data_type(infr.element_size)
            self.chan_info[i].invals =
                ffi_cast(sample_type, infr.buf)
        end
        -- create output frames and get data "pointers"
        for i = 1, #self.chan_info do
            local infr = self.chan_info[i].in_frame
            local sample_type = utils.frame_data_type(infr.element_size)
            local outfr =
                Frame_create(infr.maxsize, infr.element_size)
            outfr.hdr.ts = self.tot_count
            self.chan_info[i].out_frame = outfr
            self.chan_info[i].outvals =
                ffi_cast(sample_type, outfr.buf)
        end

        -- fill output frames, and put them in "sie buffer" pipes
        for i = 1, #self.chan_info do
            local iv = self.chan_info[i].invals
            local ov = self.chan_info[i].outvals
            for j = 0, count - 1 do
                ov[j] = iv[j]
            end
            self.chan_info[i].out_frame.count = count
            -- put output frame in output pipe(s)
            if not utils.put_frame(self.chan_info[i].out_frame,
                self.chan_info[i].pipe_list) then
                Frame_delete(self.chan_info[i].out_frame)
            end
        end

        self.tot_count = self.tot_count + count

        -- don't need the processed input frames any longer
        for i = 1, #self.chan_info do
            Frame_delete(self.chan_info[i].in_frame)
        end

        -- get a new set of input frames to process
        have_all_frames = self:get_frames()
    end
end

local m = {}

-- Creates a new instantiation of this "module".
-- Standard parameters are:
--   stanza --> decoded JSON structure containing the metadata
--       for an instance of this module
--   node_list --> list of all nodes (i.e., module instances, proxy
--       nodes) created to this point, indexed by channel name
--   first_sie_index --> NOT USED
--   meta_list --> list of all created-to-this-point nodes' stanzas,
--       indexed by channel name
-- Return values:
--   the new instance object
function m:new(stanza, node_list, first_sie_index, meta_list)
    assert(stanza, "Missing parameter 'stanza'")
    assert(node_list, "Missing parameter 'node_list'")
    assert(first_sie_index, "Missing parameter 'first_sie_index'")
    assert(meta_list, "Missing parameter 'meta_list'")
    local time_history = setmetatable({}, {__index = mt})
    time_history.stanza = stanza
    -- save a needed system preference
    time_history.prefix_dm_name = stanza.summary.include_dm_name_and_at
    -- don't need to call set_common_meta() since there are
    -- no modules downstream
    time_history.chan_info = {}
    -- validate and process needed metadata
    local input_channels = stanza.input_channels
    assert(#input_channels >= 1, "Time History data mode " ..
        "must have at least one input channel")
    --     node name
    time_history.name = stanza.name
    assert(time_history.name, "Missing node name")
    --     sample rate
    for i = 1, #input_channels do
        local node_meta = meta_list[input_channels[i]]
        assert(node_meta, "metadata stanza for Input Channel \"" ..
            tostring(input_channels[i]) .. "\" not found")
        time_history.sample_rate = node_meta.sample_rate
        if time_history.sample_rate then
            time_history.sample_rate =
                rational:new(time_history.sample_rate):tonumber()
            if time_history.sample_rate > 0 then
                break
            end
        end
    end
    assert(time_history.sample_rate and time_history.sample_rate > 0,
        string.format("Invalid input sample rate: %s",
        tostring(time_history.sample_rate)))
    --     trigger
    local trigger_mode = stanza.trigger_mode
    assert(trigger_mode, "Missing trigger mode")
    if trigger_mode == "always_on" then
        --     create input pipe structures
        local input_node
        local node_meta
        for i = 1, #input_channels do
            input_node = node_list[input_channels[i]]
            assert(input_node, "Node for Input Channel \"" ..
                tostring(input_channels[i]) .. "\" not found")
            node_meta = meta_list[input_channels[i]]
            assert(node_meta, "metadata stanza for Input Channel \"" ..
                tostring(input_channels[i]) .. "\" not found")
            local data_type = node_meta.data_type
            assert(data_type, "Missing data_type for \"" ..
                tostring(input_channels[i]) .. "\"")
            time_history.chan_info[i] =
                {name = input_channels[i],
                 --in_node = input_node,
                 in_pipe = Pipe_attach_pipe(input_node.pipe_list),
                 pipe_list = Pipe_new_pipe_list(),
                 stanza = node_meta,
                 data_type = data_type,
                 sie_buf_info = {
                    idx = stanza.sie_buf_info.idx + i - 1,
                    id = stanza.sie_buf_info.id + i - 1,
                    group = stanza.sie_buf_info.group + i - 1,
                 },
                }
            assert(time_history.chan_info[i].in_pipe and
                time_history.chan_info[i].in_pipe ~= nil,
                "could not create input pipe")
            assert(time_history.chan_info[i].pipe_list and
                time_history.chan_info[i].pipe_list ~= nil,
                "could not create pipe list")
        end
    else
        --     create trigger module instance
        local trig_chan = stanza.trigger_channel
        assert(trig_chan, "Trigger mode requires a trigger channel")
        local ok, mod =
            pcall(require, 'test_engine/modules/trigger/' .. trigger_mode)
        if not ok then
            error(string.format(__'Could not load trigger ' ..
                 'module %s (%s)', stanza.trigger_mode, stanza.name))
        end
        local trigger_module =
            mod:new(input_channels, trig_chan, node_list, meta_list)
        if not trigger_module then
            error(string.format(__'Unable to create type %s (%s)',
                'test_engine/modules/trigger/' .. trigger_mode,
                stanza.name))
        end
        assert(#trigger_module.chan_info ==
               #input_channels, "Number of trigger output pipes " ..
               "does not equal number of input channels")
        --     create input pipe structures using
        --     trigger module output pipes as this instance's
        --     input pipes
        --local input_node
        for i = 1, #input_channels do
            --input_node = node_list[input_channels[i]]
            --assert(input_node, "Node for Input Channel \"" ..
            --    tostring(input_channels[i]) .. "\" not found")
            local node_meta = meta_list[input_channels[i]]
            assert(node_meta, "metadata stanza for Input Channel \"" ..
                tostring(input_channels[i]) .. "\" not found")
            local data_type = node_meta.data_type
            assert(data_type, "Missing data_type for \"" ..
                tostring(input_channels[i]) .. "\"")
            time_history.chan_info[i] =
                {name = input_channels[i],
                 --in_node = input_node,
                 in_pipe = Pipe_attach_pipe(
                     trigger_module.chan_info[i].pipe_list),
                 pipe_list = Pipe_new_pipe_list(),
                 stanza = node_meta,
                 data_type = data_type,
                 sie_buf_info = {
                    idx = stanza.sie_buf_info.idx + i - 1,
                    id = stanza.sie_buf_info.id + i - 1,
                    group = stanza.sie_buf_info.group + i - 1,
                 },
                }
            assert(time_history.chan_info[i].in_pipe and
                time_history.chan_info[i].in_pipe ~= nil,
                "could not create input pipe")
            assert(time_history.chan_info[i].pipe_list and
                time_history.chan_info[i].pipe_list ~= nil,
                "could not create pipe list")
        end
        time_history.trigger_module = trigger_module
    end
 
    -- used for frames' time stamps (just a sample count for
    -- this module)
    time_history.tot_count = 0

    return time_history, #input_channels  -- 1 sie buf per input chan
end

return m
